"""iSCSI parameter management.

This module provides utilities for managing iSCSI parameters:
- Parameter validation
- Parameter negotiation

iSCSI parameters control:
- Data integrity (digests)
- Performance (burst lengths)
- Data transfer behavior
- Session characteristics
"""
#  Copyright: Contributors to the sts project
#  GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import logging
from typing import Callable, Final

logger = logging.getLogger(__name__)

# Parameter name mapping between iscsiadm and our API
# iscsiadm uses hierarchical names (node.conn[0].iscsi.*)
# We use simpler names for better usability
PARAM_MAP: Final[dict[str, str]] = {
    'HeaderDigest': 'node.conn[0].iscsi.HeaderDigest',  # CRC32C/None
    'MaxRecvDataSegmentLength': 'node.conn[0].iscsi.MaxRecvDataSegmentLength',  # Bytes
    'MaxXmitDataSegmentLength': 'node.conn[0].iscsi.MaxXmitDataSegmentLength',  # Bytes
    'MaxBurstLength': 'node.session.iscsi.MaxBurstLength',  # Bytes
    'FirstBurstLength': 'node.session.iscsi.FirstBurstLength',  # Bytes
    'ImmediateData': 'node.session.iscsi.ImmediateData',  # Yes/No
    'InitialR2T': 'node.session.iscsi.InitialR2T',  # Yes/No
}


def _verify_header_digest(target_value: str, initiator_value: str) -> str:
    """Verify HeaderDigest parameter."""
    # HeaderDigest negotiation:
    # - Both must support CRC32C for it to be used
    # - None is always supported
    # - Order of preference matters
    expected = 'CRC32C' if 'None' not in initiator_value or 'None' not in target_value else 'None'

    if 'CRC32C' not in initiator_value or 'CRC32C' not in target_value:
        expected = 'None'

    # Handle different preference orders
    if (initiator_value == 'CRC32C,None' and target_value == 'None,CRC32C') or (
        initiator_value == target_value == 'CRC32C,None'
    ):
        expected = 'CRC32C'
    elif (initiator_value == 'None,CRC32C' and target_value == 'CRC32C,None') or (
        initiator_value == target_value == 'None,CRC32C'
    ):
        expected = 'None'
    return expected


def _verify_max_recv_data_segment_length(initiator_value: str) -> str:
    """Verify MaxRecvDataSegmentLength parameter."""
    # Receiver controls its buffer size
    return initiator_value


def _verify_data_size_param(target_value: str, initiator_value: str) -> str:
    """Verify parameters like MaxXmitDataSegmentLength and MaxBurstLength."""
    # Use minimum of offered values for data sizes
    return target_value if int(target_value) < int(initiator_value) else initiator_value


def _verify_immediate_data(target_value: str, initiator_value: str) -> str:
    """Verify ImmediateData parameter."""
    # Both must support immediate data for it to be enabled
    return 'Yes' if target_value == initiator_value == 'Yes' else 'No'


def _verify_initial_r2t(target_value: str, initiator_value: str) -> str:
    """Verify InitialR2T parameter."""
    # Both must support disabled R2T for it to be disabled
    return 'No' if target_value == initiator_value == 'No' else 'Yes'


def _verify_first_burst_length(
    target_value: str, initiator_value: str, max_burst_length: tuple[str, str] | None
) -> str:
    """Verify FirstBurstLength parameter."""
    # First use minimum of offered values
    expected = target_value if int(target_value) < int(initiator_value) else initiator_value

    # Then ensure FirstBurstLength <= MaxBurstLength
    if max_burst_length:
        target_mbl, initiator_mbl = max_burst_length
        exp_max_burst = target_mbl if int(target_mbl) < int(initiator_mbl) else initiator_mbl
        if int(expected) > int(exp_max_burst):
            expected = exp_max_burst
    return expected


# ParamVerifierCallable defines a callable that takes any arguments and returns a string
ParamVerifierCallable = Callable[..., str]

_PARAM_VERIFIERS: Final[dict[str, ParamVerifierCallable]] = {
    'HeaderDigest': _verify_header_digest,
    'MaxRecvDataSegmentLength': _verify_max_recv_data_segment_length,
    'MaxXmitDataSegmentLength': _verify_data_size_param,
    'MaxBurstLength': _verify_data_size_param,
    'ImmediateData': _verify_immediate_data,
    'InitialR2T': _verify_initial_r2t,
    'FirstBurstLength': _verify_first_burst_length,
}


def verify_parameter(
    param: str,
    target_value: str,
    initiator_value: str,
    negotiated_value: str,
    max_burst_length: tuple[str, str] | None = None,
) -> bool:
    """Verify negotiated parameter value.

    Parameter negotiation follows RFC 7143 rules

    Args:
        param: Parameter name
        target_value: Target's offered value
        initiator_value: Initiator's offered value
        negotiated_value: Actually negotiated value
        max_burst_length: (target, initiator) MaxBurstLength values for FirstBurstLength validation

    Returns:
        True if negotiated value is correct, False otherwise
    """
    verifier = _PARAM_VERIFIERS.get(param)
    if not verifier:
        logger.warning(f'Unknown parameter: {param}')
        return False

    try:
        if param == 'FirstBurstLength':
            expected = verifier(target_value, initiator_value, max_burst_length)
        elif param == 'MaxRecvDataSegmentLength':
            expected = verifier(initiator_value)
        else:
            expected = verifier(target_value, initiator_value)
    except (ValueError, TypeError):
        logger.exception(f'Parameter validation error for {param}')
        return False

    if expected != negotiated_value:
        logger.warning(
            f"""
            Parameter {param} validation failed:
            target={target_value}
            initiator={initiator_value}
            negotiated={negotiated_value} expected={expected}
            """
        )
        return False

    return True
