"""Network interface management.

This module provides functionality for managing network interfaces:
- Interface discovery
- MAC address handling
- IP address management
- Interface operations
"""
#  Copyright: Contributors to the sts project
#  GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import ipaddress
import logging
import re
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar, Literal

from sts.network.errors import NetworkError
from sts.utils.cmdline import run

logger = logging.getLogger(__name__)

# Constants
MAC_ADDRESS_LENGTH = 12  # Length of MAC address without colons

# MAC address regex pattern (e.g. F0:DE:F1:0D:D3:C9)
MAC_PATTERN = re.compile(r'^(?:[0-9A-F]{2}:){5}[0-9A-F]{2}$', re.IGNORECASE)


@dataclass
class NetworkInterface:
    """Network interface representation.

    This class provides functionality for managing network interfaces:
    - Interface discovery
    - MAC address handling
    - IP address management
    - Interface operations

    Args:
        name: Interface name (optional, discovered from MAC)
        mac: MAC address (optional, discovered from name)
        driver: Driver name (optional, discovered from interface)
        pci_id: PCI ID (optional, discovered from interface)

    Example:
        ```python
        iface = NetworkInterface()  # Discovers first available interface
        iface = NetworkInterface(name='eth0')  # Discovers other values
        iface = NetworkInterface(mac='F0:DE:F1:0D:D3:C9')  # Discovers other values
        ```
    """

    # Optional parameters
    name: str | None = None
    mac: str | None = None
    driver: str | None = None
    pci_id: str | None = None

    # Class-level paths
    SYSFS_PATH: ClassVar[Path] = Path('/sys/class/net')

    def fetch_mac_for_name(self) -> str | None:
        """Fetch MAC address for the current interface name."""
        if not self.name:
            return None
        mac_result = run(f'ip -o link show {self.name}')
        if mac_result.succeeded:
            for line in mac_result.stdout.splitlines():
                if 'link/ether' in line:
                    for token in line.split():
                        if self.is_valid_mac(token):
                            return token

        return None

    def fetch_driver_for_name(self) -> str | None:
        """Fetch driver for the current interface name."""
        if not self.name:
            return None
        driver_path = self.SYSFS_PATH / self.name / 'device/driver'
        if driver_path.exists():
            return driver_path.resolve().name
        return None

    def fetch_pci_id_for_name(self) -> str | None:
        """Fetch PCI ID for the current interface name."""
        if not self.name:
            return None
        device_path = self.SYSFS_PATH / self.name
        if device_path.exists():
            link = device_path.resolve().as_posix()
            pci_match = re.search(r'([0-9a-f]{4}:[0-9a-f]{2}:[0-9a-f]{2}\.[0-9a-f])/net', link)
            if pci_match:
                return pci_match.group(1)
        return None

    def _discover_properties_if_name_known(self) -> None:
        """Discover MAC, driver, and PCI ID if self.name is known."""
        if not self.mac:
            self.mac = self.fetch_mac_for_name()
        if not self.driver:
            self.driver = self.fetch_driver_for_name()
        if not self.pci_id:
            self.pci_id = self.fetch_pci_id_for_name()

    def _discover_properties_if_mac_known(self) -> None:
        """Discover name, driver, and PCI ID if self.mac is known and name is not."""
        if not self.is_valid_mac(self.mac):  # type: ignore[arg-type]
            raise NetworkError(f'Invalid MAC format: {self.mac}')
        for interface in self.get_all_interfaces_basic_info():  # Use a lightweight version of get_all
            if interface.mac and self.mac and interface.mac.upper() == self.mac.upper():
                self.name = interface.name
                # Once name is found, discover other properties based on name
                self._discover_properties_if_name_known()
                return  # Found and populated

    def _discover_first_available_interface_name(self) -> None:
        """Discover the first available interface name if no parameters are provided."""
        result = run('ls /sys/class/net')
        if result.succeeded:
            for name_str in result.stdout.splitlines():
                if any(re.match(pattern, name_str) for pattern in self._get_skip_patterns()):
                    continue
                self.name = name_str
                break

    def __post_init__(self) -> None:
        """Initialize network interface.

        Discovers name, MAC, driver, and PCI ID if not provided.
        """
        if not self.name and not self.mac:
            self._discover_first_available_interface_name()

        if self.name:
            self._discover_properties_if_name_known()
        elif self.mac:  # self.name is None here
            self._discover_properties_if_mac_known()

        # Final validation for MAC if it was provided but no name could be found
        # (e.g. MAC provided, but doesn't exist on system)
        if self.mac and not self.name and not self.is_valid_mac(self.mac):
            raise NetworkError(f'Invalid MAC format or MAC not found: {self.mac}')

    @staticmethod
    def _get_skip_patterns() -> list[str]:
        """Get patterns for interfaces to skip.

        Returns:
            List of regex patterns
        """
        return [
            r'^lo$',  # Loopback
            r'^tun[0-9]+$',  # TUN/TAP
            r'^vboxnet[0-9]+$',  # VirtualBox
            r'\.',  # Sub-interfaces
        ]

    @staticmethod
    def is_valid_mac(mac: str) -> bool:
        """Check if MAC address is valid.

        Args:
            mac: MAC address to check

        Returns:
            True if valid, False otherwise

        Example:
            ```python
            NetworkInterface.is_valid_mac('F0:DE:F1:0D:D3:C9')
            True
            ```
        """
        return bool(MAC_PATTERN.match(mac))

    @staticmethod
    def standardize_mac(mac: str) -> str | None:
        """Standardize MAC address format.

        Args:
            mac: MAC address to standardize

        Returns:
            Standardized MAC or None if invalid

        Example:
            ```python
            NetworkInterface.standardize_mac('F0-DE-F1-0D-D3-C9')
            'F0:DE:F1:0D:D3:C9'
            ```
        """
        if not mac:
            return None

        # Remove 0x and non-hex characters
        mac = re.sub(r'^0x', '', mac.upper())
        mac = re.sub(r'[^0-9A-F]', '', mac)

        # Add : every 2 characters
        if len(mac) == MAC_ADDRESS_LENGTH:
            mac = ':'.join(mac[i : i + 2] for i in range(0, MAC_ADDRESS_LENGTH, 2))

        return mac if NetworkInterface.is_valid_mac(mac) else None

    @property
    def ip_addresses(self) -> list[str]:
        """Get IP addresses.

        Returns:
            List of IP addresses

        Example:
            ```python
            iface.ip_addresses
            ['192.168.1.100', 'fe80::1234:5678:9abc:def0']
            ```
        """
        if not self.name:
            return []

        result = run(f'ip addr show {self.name}')
        if result.failed:
            return []

        addresses = []
        for line in result.stdout.splitlines():
            if 'inet' in line:
                # Parse line like: inet 192.168.1.100/24
                try:
                    addr = line.split()[1].split('/')[0]
                    addresses.append(addr)
                except (IndexError, ValueError):
                    continue

        return addresses

    @staticmethod
    def get_ip_version(addr: str) -> Literal[4, 6] | None:
        """Get IP version.

        Args:
            addr: IP address

        Returns:
            4 for IPv4, 6 for IPv6, None if invalid

        Example:
            ```python
            iface.get_ip_version('192.168.1.100')
            4
            ```
        """
        try:
            return ipaddress.ip_address(addr).version
        except ValueError:
            return None

    def up(self) -> bool:
        """Bring interface up.

        Returns:
            True if successful, False otherwise

        Example:
            ```python
            iface.up()
            True
            ```
        """
        if not self.name:
            logger.error('Interface name required')
            return False

        result = run(f'ip link set {self.name} up')
        return result.succeeded

    def down(self) -> bool:
        """Bring interface down.

        Returns:
            True if successful, False otherwise

        Example:
            ```python
            iface.down()
            True
            ```
        """
        if not self.name:
            logger.error('Interface name required')
            return False

        result = run(f'ip link set {self.name} down')
        return result.succeeded

    @classmethod
    def get_all(cls) -> list[NetworkInterface]:
        """Get list of all network interfaces.

        Returns:
            List of NetworkInterface instances

        Example:
            ```python
            NetworkInterface.get_all()
            [NetworkInterface(name='eth0', ...), NetworkInterface(name='eth1', ...)]
            ```
        """
        interfaces: list[NetworkInterface] = []

        # Get all interfaces
        result = run('ls /sys/class/net')
        if result.failed:
            return interfaces

        for name_str in result.stdout.splitlines():
            # Skip special interfaces
            if any(re.match(pattern, name_str) for pattern in cls._get_skip_patterns()):
                continue

            # Create interface
            try:
                # Pass only name to avoid recursion in __post_init__ during get_all
                interface = cls(name=name_str)
                # Manually trigger discovery for mac, driver, pci_id for this instance
                if not interface.mac:
                    interface.mac = interface.fetch_mac_for_name()
                if not interface.driver:
                    interface.driver = interface.fetch_driver_for_name()
                if not interface.pci_id:
                    interface.pci_id = interface.fetch_pci_id_for_name()

                if interface.mac:  # Only add if MAC was discovered
                    interfaces.append(interface)
            except (OSError, ValueError, NetworkError) as e:
                logger.warning(f'Failed to create interface {name_str}: {e}')
        return interfaces

    @classmethod
    def get_all_interfaces_basic_info(cls) -> list[NetworkInterface]:
        """Get list of all network interfaces with basic MAC and name."""
        # This is a helper for _discover_properties_if_mac_known to avoid deep recursion.
        # It only fetches name and MAC.
        interfaces: list[NetworkInterface] = []
        result = run('ls /sys/class/net')
        if result.failed:
            return interfaces

        for name_str in result.stdout.splitlines():
            if any(re.match(pattern, name_str) for pattern in cls._get_skip_patterns()):
                continue

            temp_iface = cls.__new__(cls)  # Create instance without calling __init__ / __post_init__
            temp_iface.name = name_str
            temp_iface.mac = None
            temp_iface.driver = None
            temp_iface.pci_id = None

            mac_addr = temp_iface.fetch_mac_for_name()
            if mac_addr:
                temp_iface.mac = mac_addr
                interfaces.append(temp_iface)
        return interfaces

    @classmethod
    def get_by_mac(cls, mac: str) -> NetworkInterface | None:
        """Get interface by MAC address.

        Args:
            mac: MAC address

        Returns:
            NetworkInterface instance or None if not found

        Example:
            ```python
            NetworkInterface.get_by_mac('F0:DE:F1:0D:D3:C9')
            NetworkInterface(name='eth0', ...)
            ```
        """
        std_mac = cls.standardize_mac(mac)
        if not std_mac:
            return None

        try:
            return cls(mac=std_mac)
        except NetworkError:
            return None

    @classmethod
    def get_by_ip(cls, ip: str) -> NetworkInterface | None:
        """Get interface by IP address.

        Args:
            ip: IP address

        Returns:
            NetworkInterface instance or None if not found

        Example:
            ```python
            NetworkInterface.get_by_ip('192.168.1.100')
            NetworkInterface(name='eth0', ...)
            ```
        """
        for interface in cls.get_all():
            if ip in interface.ip_addresses:
                return interface

        return None
