"""Process management.

This module provides functionality for managing system processes:
- Process information
- Process control
- Process monitoring
"""
#  Copyright: Contributors to the sts project
#  GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import logging
import os
import signal
import time
from dataclasses import dataclass
from pathlib import Path

from sts.utils.cmdline import run

logger = logging.getLogger(__name__)


@dataclass
class ProcessInfo:
    """Process information.

    This class provides functionality for managing process information:
    - Process metadata
    - Process status
    - Process control

    Args:
        pid: Process ID (optional, discovered from name)
        name: Process name (optional, discovered from pid)
        status: Process status (optional, discovered from process)
        cmdline: Process command line (optional, discovered from process)

    Example:
        ```python
        info = ProcessInfo()  # Discovers first available process
        info = ProcessInfo(pid=1234)  # Discovers other values
        info = ProcessInfo(name='sleep')  # Discovers matching process
        ```
    """

    # Optional parameters
    pid: int | None = None
    name: str | None = None
    status: str | None = None
    cmdline: str | None = None

    @staticmethod
    def _get_proc_entries() -> list[Path]:
        """Get all directory entries in /proc."""
        try:
            return [entry for entry in Path('/proc').iterdir() if entry.is_dir()]
        except OSError:
            logger.warning('Failed to list /proc entries')
            return []

    def _find_first_available_pid(self) -> None:
        """Find the PID of the first available process if no PID/name is provided."""
        for entry in ProcessInfo._get_proc_entries():
            if entry.name.isdigit():
                self.pid = int(entry.name)
                return

    def _find_pid_for_name(self) -> None:
        """Find PID for the current self.name if PID is not provided."""
        if not self.name:
            return
        for entry in ProcessInfo._get_proc_entries():
            if not entry.name.isdigit():
                continue
            comm_path = entry / 'comm'
            try:
                if comm_path.exists() and comm_path.read_text().strip() == self.name:
                    self.pid = int(entry.name)
                    return
            except (OSError, ValueError):
                logger.debug(f'Error reading comm for {entry.name}', exc_info=True)
                continue

    @staticmethod
    def _read_comm_file(proc_pid_path: Path) -> str | None:
        """Read process name from /proc/[pid]/comm."""
        comm_path = proc_pid_path / 'comm'
        if comm_path.exists():
            return comm_path.read_text().strip()
        return None

    @staticmethod
    def _read_status_file(proc_pid_path: Path) -> str | None:
        """Read process state from /proc/[pid]/status."""
        status_path = proc_pid_path / 'status'
        if status_path.exists():
            status_content = status_path.read_text().splitlines()
            for line in status_content:
                if line.startswith('State:'):
                    return line.split(':', 1)[1].strip()
        return None

    @staticmethod
    def _read_cmdline_file(proc_pid_path: Path) -> str | None:
        """Read command line from /proc/[pid]/cmdline."""
        cmdline_path = proc_pid_path / 'cmdline'
        if cmdline_path.exists():
            try:
                raw = cmdline_path.read_bytes()
            except OSError:
                return None
            # Replace NULs with spaces and decode with fallback
            return raw.replace(b'\x00', b' ').decode(errors='replace').strip()
        return None

    def _populate_info_from_pid(self) -> None:
        """Populate name, status, and cmdline if self.pid is known."""
        if not self.pid:
            return

        proc_pid_path = Path('/proc') / str(self.pid)
        if not proc_pid_path.exists():
            logger.warning(f'Process path {proc_pid_path} does not exist.')
            self.pid = None  # PID is invalid
            return

        try:
            if not self.name:
                self.name = ProcessInfo._read_comm_file(proc_pid_path)

            # Always try to get current status and cmdline
            self.status = ProcessInfo._read_status_file(proc_pid_path)
            self.cmdline = ProcessInfo._read_cmdline_file(proc_pid_path)

        except OSError:
            logger.exception(f'Failed to get process info for PID {self.pid}')
            # Potentially clear pid if info is critical and unreadable
            # For now, we allow partial info if some reads fail

    def __post_init__(self) -> None:
        """Initialize process information.

        Discovers process information based on provided parameters.
        """
        if not self.pid and not self.name:
            self._find_first_available_pid()
        elif self.name and not self.pid:
            self._find_pid_for_name()

        if self.pid:
            self._populate_info_from_pid()
        elif not self.name:  # Case where no PID was found and no name was given
            logger.debug('Could not determine PID and no name provided.')

    @property
    def exists(self) -> bool:
        """Check if process exists.

        Returns:
            True if exists, False otherwise

        Example:
            ```python
            info = ProcessInfo(pid=1234)
            info.exists
            True
            ```
        """
        return bool(self.pid and Path('/proc', str(self.pid)).exists())

    @property
    def running(self) -> bool:
        """Check if process is running.

        Returns:
            True if running, False otherwise

        Example:
            ```python
            info = ProcessInfo(pid=1234)
            info.running
            True
            ```
        """
        if not self.pid:
            return False

        try:
            os.kill(self.pid, 0)
        except ProcessLookupError:
            return False
        except PermissionError:
            # Process exists but we don't have permission to send signals
            return True
        return True

    def kill(self, timeout: float = 1.0) -> bool:
        """Kill process.

        Args:
            timeout: Timeout in seconds between SIGTERM and SIGKILL

        Returns:
            True if successful, False otherwise

        Example:
            ```python
            info = ProcessInfo(pid=1234)
            info.kill()
            True
            ```
        """
        if not self.pid:
            return False

        if not self.running:
            return True

        try:
            # Try SIGTERM first
            os.kill(self.pid, signal.SIGTERM)
            time.sleep(timeout)

            # If still running, use SIGKILL
            if self.running:
                os.kill(self.pid, signal.SIGKILL)
                time.sleep(timeout)

                # Check if process is still running
                if self.running:
                    logger.error(f'Failed to kill process {self.pid}')
                    return False

        except ProcessLookupError:
            # Process already terminated
            pass
        except PermissionError:
            logger.exception(f'Permission denied killing process {self.pid}')
            return False

        return True

    @classmethod
    def from_pid(cls, pid: int) -> ProcessInfo | None:
        """Get process information by PID.

        Args:
            pid: Process ID

        Returns:
            Process information or None if not found

        Example:
            ```python
            info = ProcessInfo.from_pid(1234)
            info.status
            'running'
            ```
        """
        info = cls(pid=pid)
        return info if info.exists else None

    @classmethod
    def from_name(cls, name: str) -> ProcessInfo | None:
        """Get process information by name.

        Args:
            name: Process name

        Returns:
            Process information or None if not found

        Example:
            ```python
            info = ProcessInfo.from_name('sleep')
            info.pid
            1234
            ```
        """
        info = cls(name=name)
        return info if info.exists else None


class ProcessManager:
    """Process manager functionality.

    This class provides functionality for managing system processes:
    - Process control
    - Process monitoring
    - Process information

    Example:
        ```python
        pm = ProcessManager()
        pm.kill_all('sleep')
        True
        ```
    """

    def __init__(self) -> None:
        """Initialize process manager."""
        self.proc_path = Path('/proc')

    def get_all(self) -> list[ProcessInfo]:
        """Get list of all processes.

        Returns:
            List of process information

        Example:
            ```python
            pm = ProcessManager()
            pm.get_all()
            [ProcessInfo(pid=1, ...), ProcessInfo(pid=2, ...)]
            ```
        """
        processes = []

        try:
            # Get list of process directories
            proc_entries = [entry for entry in self.proc_path.iterdir() if entry.is_dir()]

            # Filter numeric directories and create processes
            for entry in proc_entries:
                if entry.name.isdigit():
                    pid = int(entry.name)
                    if info := ProcessInfo.from_pid(pid):
                        processes.append(info)

        except OSError:
            logger.exception('Failed to get process list')
            return []

        return processes

    def get_by_name(self, name: str) -> list[ProcessInfo]:
        """Get processes by name.

        Args:
            name: Process name

        Returns:
            List of matching processes

        Example:
            ```python
            pm = ProcessManager()
            pm.get_by_name('sleep')
            [ProcessInfo(pid=1234, name='sleep', ...)]
            ```
        """
        return [p for p in self.get_all() if p.name == name]

    def kill_all(self, name: str, timeout: float = 1.0) -> bool:
        """Kill all processes by name.

        Args:
            name: Process name
            timeout: Timeout in seconds between SIGTERM and SIGKILL

        Returns:
            True if successful, False otherwise

        Example:
            ```python
            pm = ProcessManager()
            pm.kill_all('sleep')
            True
            ```
        """
        result = run(f'killall {name}')
        if result.failed:
            return False

        # Wait for processes to finish
        time.sleep(timeout)

        # Check if any processes still running
        return not bool(self.get_by_name(name))
